import os
import sys

from easyprocess import EasyProcess


def py_minor():
    return sys.version_info[1]


def platform_is_osx():
    return sys.platform == "darwin"


def platform_is_win():
    return sys.platform == "win32"


def platform_is_linux():
    return sys.platform.startswith("linux")


def use_x_display():
    if platform_is_win():
        return False
    if platform_is_osx():
        return False
    DISPLAY = os.environ.get("DISPLAY")
    XDG_SESSION_TYPE = os.environ.get("XDG_SESSION_TYPE")
    # Xwayland can not be used for screenshot
    return DISPLAY and XDG_SESSION_TYPE != "wayland"


def extract_version(txt):
    """This function tries to extract the version from the help text of any
    program."""
    words = txt.replace(",", " ").split()
    version = None
    for x in reversed(words):
        if len(x) > 2:
            if x[0].lower() == "v":
                x = x[1:]
            if "." in x and x[0].isdigit():
                version = x
                break
    return version


def run_mod_as_subproc(name, params=[]):
    python = sys.executable
    cmd = [python, "-m", name] + params
    p = EasyProcess(cmd).call()
    return p



from bt import backend_to_check, prog_check

from pyscreenshot.util import use_x_display

if use_x_display():
    if prog_check(["import", "-version"]):

        def test_imagemagick():
            backend_to_check("imagemagick")



import logging
import os

from PIL import Image

from pyscreenshot.plugins.backend import UNKNOWN_VERSION, CBackend

log = logging.getLogger(__name__)


class KdeDBusError(Exception):
    pass


# https://gitlab.gnome.org/GNOME/gimp/-/issues/6626
# The org.kde.kwin.Screenshot interface is deprecated in KDE Plasma 5.22.

# "The process is not authorized to take a screenshot"


class KwinDBusWrapper(CBackend):
    name = "kwin_dbus"
    is_subprocess = True

    def grab(self, bbox=None):
        has_jeepney = False
        try:
            # from jeepney import new_method_call
            from jeepney.io.blocking import open_dbus_connection  # type: ignore
            from jeepney.wrappers import MessageGenerator  # type: ignore
            from jeepney.wrappers import new_method_call

            has_jeepney = True
        except ImportError:
            pass

        if not has_jeepney:
            raise KdeDBusError("jeepney library is missing")

        class Screenshot(MessageGenerator):
            interface = "org.kde.kwin.Screenshot"

            def __init__(self, object_path="/Screenshot", bus_name="org.kde.KWin"):
                super().__init__(object_path=object_path, bus_name=bus_name)

            def screenshotFullscreen(self, captureCursor):
                return new_method_call(
                    self, "screenshotFullscreen", "b", (captureCursor,)
                )

            def screenshotArea(self, x, y, width, height, captureCursor):
                return new_method_call(
                    self,
                    "screenshotArea",
                    "iiiib",
                    (x, y, width, height, captureCursor),
                )

        # https://jeepney.readthedocs.io/en/latest/integrate.html
        connection = open_dbus_connection(bus="SESSION")
        dbscr = Screenshot()
        # bbox not working:
        #   if bbox: msg = dbscr.screenshotArea(bbox[0], bbox[1], bbox[2] - bbox[0], bbox[3] - bbox[1], False)
        msg = dbscr.screenshotFullscreen(False)
        reply = connection.send_and_get_reply(msg)
        filename = reply.body[0]
        if not filename:
            raise KdeDBusError()

        im = Image.open(filename)
        os.remove(filename)
        if bbox:
            im = im.crop(bbox)
        return im

    def backend_version(self):
        return UNKNOWN_VERSION



import logging

from entrypoint2 import entrypoint
from PIL import Image

# from size import display_size

log = logging.getLogger(__name__)


def generate_image(w, h):
    # w, h = display_size()
    log.debug("display size: %s x %s", w, h)
    if w <= 0 or h <= 0:
        raise ValueError("invalid display size %s x %s" % (w, h))
    img = Image.new("RGB", (w, h), "black")  # Create a new black image
    pixels = img.load()  # Create the pixel map
    i = 0
    B = 42
    for x in range(img.size[0]):  # For every pixel:
        for y in range(img.size[1]):
            r = int(x * 255 / w)
            g = int(y * 255 / h)

            b = int(((x % B) * 255 / B + (y % B) * 255 / B) / 2)
            pixels[x, y] = (r, g, b)  # Set the colour accordingly
            i += 1
    return img


@entrypoint
def main():
    im = generate_image()
    im.show()



from easyprocess import EasyProcess

from pyscreenshot.plugins.backend import CBackend
from pyscreenshot.tempexport import read_prog_img
from pyscreenshot.util import extract_version, platform_is_osx

PROGRAM = "import"
# http://www.imagemagick.org/


class ImagemagickBackendError(Exception):
    pass


class ImagemagickWrapper(CBackend):
    name = "imagemagick"
    is_subprocess = True

    def grab(self, bbox=None):
        if platform_is_osx():
            raise ImagemagickBackendError("osx not supported")

        command = [PROGRAM, "-silent", "-window", "root"]
        if bbox:
            pbox = "{}x{}+{}+{}".format(
                bbox[2] - bbox[0], bbox[3] - bbox[1], bbox[0], bbox[1]
            )
            command += ["-crop", pbox]
        im = read_prog_img(command)
        return im

    def backend_version(self):
        stdout = EasyProcess([PROGRAM, "-version"]).call().stdout
        s = stdout.splitlines()[0]
        return extract_version(s.replace("-", " "))



import atexit
import logging
import os
import shutil
import tempfile
from logging import DEBUG

log = logging.getLogger(__name__)

CLEANUP = False

# for image debug
USE_TESTOUT_DIR = 1


class ImageDebug(object):
    def __init__(self):
        self.index = 0
        d = None
        if USE_TESTOUT_DIR:
            d = os.path.join(os.getcwd(), "testout")
            os.makedirs(d, exist_ok=True)
        self.dir = tempfile.mkdtemp(dir=d, prefix="pyscreenshot_img_")
        if CLEANUP:
            atexit.register(shutil.rmtree, self.dir)

    def img_debug(self, im, text):
        if not log.isEnabledFor(DEBUG):
            return
        fname = os.path.join(self.dir, str(self.index).zfill(3) + "_" + text + ".png")
        im.save(fname)
        log.debug("image (%s) was saved:" % im + fname)
        self.index += 1



"""
import logging
import tempfile
import tkinter
from os.path import join

from entrypoint2 import entrypoint
from PIL import Image, ImageTk

from genimg import generate_image
from pyscreenshot.util import platform_is_osx

log = logging.getLogger(__name__)


# https://stackoverflow.com/questions/47316266/can-i-display-image-in-full-screen-mode-with-pil/47317411
def fillscreen_tk(fimage):
    pilImage = Image.open(fimage)
    root = tkinter.Tk()
    w, h = root.winfo_screenwidth(), root.winfo_screenheight()

    root.config(cursor="none")
    root.geometry("%dx%d+0+0" % (w, h))
    root.focus_set()
    root.bind("<Escape>", lambda e: (e.widget.withdraw(), e.widget.quit()))
    root.bind("q", lambda e: (e.widget.withdraw(), e.widget.quit()))

    # print(w,h)
    canvas = tkinter.Canvas(root, width=w, height=h)
    canvas.pack()
    canvas.configure(background="red")
    assert pilImage.size == (w, h)
    image = ImageTk.PhotoImage(pilImage)
    imagesprite = canvas.create_image(w / 2, h / 2, image=image)

    # make fullscreen
    if platform_is_osx():
        root.tk.call(
            "::tk::unsupported::MacWindowStyle", "style", root._w, "plain", "none"
        )
    root.attributes("-fullscreen", True)
    # root.overrideredirect(True)
    root.attributes("-topmost", True)

    root.mainloop()


@entrypoint
def main(image=""):
    if not image:
        d = tempfile.mkdtemp(prefix="fillscreen")
        image = join(d, "ref.png")
        im = generate_image()
        im.save(image)
    fillscreen_tk(image)
"""



import logging

from easyprocess import EasyProcess

from pyscreenshot.plugins.backend import CBackend
from pyscreenshot.tempexport import read_prog_img
from pyscreenshot.util import extract_version

log = logging.getLogger(__name__)


PROGRAM = "scrot"


class ScrotWrapper(CBackend):
    name = "scrot"
    is_subprocess = True

    def grab(self, bbox=None):
        im = read_prog_img([PROGRAM, "--silent"])
        if bbox:
            im = im.crop(bbox)
        return im

    def backend_version(self):
        return extract_version(EasyProcess([PROGRAM, "-version"]).call().stdout)



import logging
import os
import tempfile
from os.path import join

import pygame
from entrypoint2 import entrypoint
from genimg import generate_image
from pygame.locals import K_ESCAPE, KEYDOWN

log = logging.getLogger(__name__)


def fillscreen_pygame(fimage):
    # DBus org.kde.kwin.Screenshot disappears on Kubuntu 20.04 after starting pygame
    #  fix: System Settings > Compositor > uncheck Allow apps to turn off compositing.
    #       or https://www.pygame.org/docs/ref/pygame.html
    #           SDL_VIDEO_X11_NET_WM_BYPASS_COMPOSITOR
    #           Set to "0" to re-enable the compositor.
    os.environ["SDL_VIDEO_X11_NET_WM_BYPASS_COMPOSITOR"] = "0"
    pygame.init()
    pygame.mouse.set_visible(0)
    log.info("pygame modes:%s", pygame.display.list_modes())
    # log.info("pygame info:%s", pygame.display.Info())
    log.info("env $DISPLAY= %s", os.environ.get("DISPLAY"))

    infoObject = pygame.display.Info()
    im = generate_image(infoObject.current_w, infoObject.current_h)
    im.save(fimage)

    windowSurface = pygame.display.set_mode((0, 0), pygame.FULLSCREEN)
    img = pygame.image.load(fimage)

    mainLoop = True

    while mainLoop:
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                mainLoop = False
            elif event.type == KEYDOWN:
                if event.key == K_ESCAPE:
                    mainLoop = False

        windowSurface.blit(img, (0, 0))
        # pygame.display.flip()
        pygame.display.update()

    pygame.quit()


@entrypoint
def main(image=""):
    if not image:
        d = tempfile.mkdtemp(prefix="fillscreen")
        # atexit.register(lambda: rmtree(d))
        image = join(d, "ref.bmp")
    # im = generate_image()
    # im.save(image)
    fillscreen_pygame(image)



from entrypoint2 import entrypoint

import pyscreenshot
from pyscreenshot.imcodec import codec


@entrypoint
def main(filename="", bbox="", backend="", show=False):
    """Copy the contents of the screen to file.

    :param filename: output file
    :param show: show image
    :param bbox: bounding box coordinates x1:y1:x2:y2
    :param backend: back-end can be forced if set (example:scrot, wx,..),
                    otherwise back-end is automatic
    """
    backend = backend if backend else None
    if bbox:
        x1, y1, x2, y2 = map(int, bbox.split(":"))
        bbox = x1, y1, x2, y2
    else:
        bbox = None

    im = pyscreenshot.grab(bbox=bbox, childprocess=False, backend=backend)
    if filename:
        b = codec[0](im)
        with open(filename, "wb") as f:
            f.write(b)
    if show:
        im.show()



from bt import backend_to_check, check_import

from pyscreenshot.util import platform_is_osx

# qt color problem on osx
if not platform_is_osx():
    if check_import("PySide"):

        def test_pyside():
            backend_to_check("pyside")



import os
import sys

from easyprocess import EasyProcess
from entrypoint2 import entrypoint
from pyvirtualdisplay import Display


def check_double_disp(backend):
    python = sys.executable
    if not backend:
        backend = ""

    # the main process may crash,
    # so it must be tested in subprocess
    fname = __file__.rsplit(".", 1)[0] + ".py"
    cmd = [python, fname, backend, "--debug"]
    p = EasyProcess(cmd).call()
    assert p.return_code == 0


@entrypoint
def main(backend):
    if not backend:
        backend = None

    sys.path.append(os.getcwd())
    from pyscreenshot import grab

    with Display(visible=False):
        img = grab(backend=backend)
        assert img

    with Display(visible=False):
        img = grab(backend=backend)
        assert img



from bt import backend_to_check, check_import

from pyscreenshot.util import use_x_display

ok = False
if check_import("gi"):
    if use_x_display():
        import gi

        # Arch: AttributeError: module 'gi' has no attribute 'require_version'
        try:
            gi.require_version("Gdk", "3.0")
            ok = True
        except AttributeError:
            pass
        except ValueError:
            pass
if ok:

    def test_pygdk3():
        backend_to_check("pygdk3")



# -*- coding: utf-8 -*-
"""Gdk3-based screenshotting.

Adapted from https://stackoverflow.com/a/37768950/81636, but uses
buffers directly instead of saving intermediate files (which is slow).
"""

from PIL import Image

from pyscreenshot.plugins.backend import CBackend
from pyscreenshot.util import platform_is_osx


class Gdk3BackendError(Exception):
    pass


class Gdk3PixbufWrapper(CBackend):
    name = "pygdk3"

    def grab(self, bbox=None):
        """Grabs an image directly to a buffer.

        :param bbox: Optional tuple or list containing (x1, y1, x2, y2) coordinates
            of sub-region to capture.
        :return: PIL RGB image
        :raises: ValueError, if image data does not have 3 channels (RGB), each with 8
            bits.
        :rtype: Image
        """
        if platform_is_osx():
            raise Gdk3BackendError("osx not supported")
        import gi  # type: ignore

        gi.require_version("Gdk", "3.0")
        # gi.require_version('GdkPixbuf', '2.0')
        from gi.repository import Gdk, GdkPixbuf  # type: ignore

        # read_pixel_bytes: New in version 2.32.
        if GdkPixbuf.PIXBUF_MAJOR == 2:
            if GdkPixbuf.PIXBUF_MINOR < 32:
                raise ValueError(
                    "GdkPixbuf min supported version: 2.32   current:"
                    + GdkPixbuf.PIXBUF_VERSION
                )

        w = Gdk.get_default_root_window()
        if bbox is not None:
            g = [bbox[0], bbox[1], bbox[2] - bbox[0], bbox[3] - bbox[1]]
        else:
            g = w.get_geometry()
        pb = Gdk.pixbuf_get_from_window(w, *g)
        if not pb:
            raise Gdk3BackendError("empty buffer")

        if pb.get_bits_per_sample() != 8:
            raise Gdk3BackendError("Expected 8 bits per pixel.")
        elif pb.get_n_channels() != 3:
            raise Gdk3BackendError("Expected RGB image.")

        # Read the entire buffer into a python bytes object.
        # read_pixel_bytes: New in version 2.32.
        pixel_bytes = pb.read_pixel_bytes().get_data()  # type: bytes
        width, height = g[2], g[3]

        # Probably for SSE alignment reasons, the pixbuf has extra data in each line.
        # The args after "raw" help handle this; see
        # http://effbot.org/imagingbook/decoder.htm#the-raw-decoder
        return Image.frombytes(
            "RGB", (width, height), pixel_bytes, "raw", "RGB", pb.get_rowstride(), 1
        )

    def backend_version(self):
        import gi

        return ".".join(map(str, gi.version_info))



from bt import backend_to_check, gnome, kde, prog_check

# 'gnome-screenshot' makes strange effects on screen.
# It can disturb later tests.
# Make this test the last by starting the name with 'x'
if gnome():
    if prog_check(["gnome-screenshot", "--version"]):

        def test_gnome_screenshot():
            assert not kde()
            # the flash effect can be seen on the next screenshot,
            # so some delay is needed
            backend_to_check("gnome-screenshot", delay=1)



import logging
import os

from PIL import Image

from pyscreenshot.plugins.backend import UNKNOWN_VERSION, CBackend

log = logging.getLogger(__name__)


class FreedesktopDBusError(Exception):
    pass


class FreedesktopDBusWrapper(CBackend):
    name = "freedesktop_dbus"
    is_subprocess = True

    def grab(self, bbox=None):
        has_jeepney = False
        try:
            from jeepney import DBusAddress, new_method_call
            from jeepney.bus_messages import MatchRule, message_bus
            from jeepney.io.blocking import Proxy, open_dbus_connection

            has_jeepney = True
        except ImportError:
            pass

        if not has_jeepney:
            raise FreedesktopDBusError("jeepney library is missing")

        portal = DBusAddress(
            object_path="/org/freedesktop/portal/desktop",
            bus_name="org.freedesktop.portal.Desktop",
        )
        screenshot = portal.with_interface("org.freedesktop.portal.Screenshot")

        conn = open_dbus_connection()

        token = "pyscreenshot"
        sender_name = conn.unique_name[1:].replace(".", "_")
        handle = f"/org/freedesktop/portal/desktop/request/{sender_name}/{token}"

        response_rule = MatchRule(
            type="signal", interface="org.freedesktop.portal.Request", path=handle
        )
        Proxy(message_bus, conn).AddMatch(response_rule)

        with conn.filter(response_rule) as responses:
            req = new_method_call(
                screenshot,
                "Screenshot",
                "sa{sv}",
                ("", {"handle_token": ("s", token), "interactive": ("b", False)}),
            )
            conn.send_and_get_reply(req)
            response_msg = conn.recv_until_filtered(responses)

        response, results = response_msg.body

        im = False
        if response == 0:
            filename = results["uri"][1].split("file://", 1)[-1]
            if os.path.isfile(filename):
                im = Image.open(filename)
                os.remove(filename)

        conn.close()

        if bbox and im:
            im = im.crop(bbox)
        return im

    def backend_version(self):
        return UNKNOWN_VERSION



import io
import logging

from PIL import Image

from pyscreenshot.plugins.backend import CBackend

log = logging.getLogger(__name__)

# based on qt4 backend

app = None


class PySideGrabWindow(CBackend):
    name = "pyside"

    def grab_to_buffer(self, buff, file_type="png"):
        from PySide import QtCore, QtGui  # type: ignore

        QApplication = QtGui.QApplication
        QBuffer = QtCore.QBuffer
        QIODevice = QtCore.QIODevice
        QPixmap = QtGui.QPixmap

        global app
        if not app:
            app = QApplication([])
        qbuffer = QBuffer()
        qbuffer.open(QIODevice.ReadWrite)
        QPixmap.grabWindow(QApplication.desktop().winId()).save(qbuffer, file_type)
        # https://stackoverflow.com/questions/52291585/pyside2-typeerror-bytes-object-cannot-be-interpreted-as-an-integer
        buff.write(qbuffer.data().data())
        qbuffer.close()

    def grab(self, bbox=None):
        strio = io.BytesIO()
        self.grab_to_buffer(strio)
        strio.seek(0)
        im = Image.open(strio)
        if bbox:
            im = im.crop(bbox)
        return im

    def backend_version(self):
        import PySide

        return PySide.__version__



import os
from tempfile import TemporaryDirectory

from bt import prog_check

from pyscreenshot.util import run_mod_as_subproc


def check_example(name):
    owd = os.getcwd()
    with TemporaryDirectory(prefix="pyscreenshot") as tmpdirname:
        try:
            os.chdir(tmpdirname)
            assert run_mod_as_subproc("pyscreenshot.examples." + name).return_code == 0
        finally:
            os.chdir(owd)


def test_grabbox():
    check_example("grabbox")


def test_grabfullscreen():
    check_example("grabfullscreen")


if prog_check(["Xvfb", "-version"]):

    def test_virtdisp():
        check_example("virtdisp")



# Javier Escalada Gomez
#
# from:
# https://stackoverflow.com/questions/4524723/take-screenshot-in-python-on-mac-os-x

from pyscreenshot.plugins.backend import CBackend
from pyscreenshot.tempexport import read_func_img


class MacQuartzWrapper(CBackend):
    name = "mac_quartz"

    def grab(self, bbox=None):
        im = read_func_img(self._grab_to_file, bbox)
        return im

    def _grab_to_file(self, filename, bbox=None, dpi=72):
        # Should query dpi from somewhere, e.g for retina displays?
        import LaunchServices  # type: ignore
        import Quartz  # type: ignore
        import Quartz.CoreGraphics as CG  # type: ignore
        from Cocoa import NSURL  # type: ignore

        if bbox:
            width = bbox[2] - bbox[0]
            height = bbox[3] - bbox[1]
            region = CG.CGRectMake(bbox[0], bbox[1], width, height)
        else:
            region = CG.CGRectInfinite

        # Create screenshot as CGImage
        image = CG.CGWindowListCreateImage(
            region,
            CG.kCGWindowListOptionOnScreenOnly,
            CG.kCGNullWindowID,
            CG.kCGWindowImageDefault,
        )

        file_type = LaunchServices.kUTTypePNG

        url = NSURL.fileURLWithPath_(filename)

        dest = Quartz.CGImageDestinationCreateWithURL(
            url,
            file_type,
            # 1 image in file
            1,
            None,
        )

        properties = {
            Quartz.kCGImagePropertyDPIWidth: dpi,
            Quartz.kCGImagePropertyDPIHeight: dpi,
        }

        # Add the image to the destination, characterizing the image with
        # the properties dictionary.
        Quartz.CGImageDestinationAddImage(dest, image, properties)

        # When all the images (only 1 in this example) are added to the destination,
        # finalize the CGImageDestination object.
        Quartz.CGImageDestinationFinalize(dest)

    def backend_version(self):
        import objc  # type: ignore

        return objc.__version__



import os

from bt import backend_to_check

PYSCREENSHOT_TEST_FREEDESKTOP_DBUS = os.environ.get(
    "PYSCREENSHOT_TEST_FREEDESKTOP_DBUS"
)

if PYSCREENSHOT_TEST_FREEDESKTOP_DBUS:

    def test_freedesktop_dbus():
        # the previous confirmation dialog can be seen on the next screenshot,
        # so some delay is needed
        # 0.1 is not enough
        backend_to_check("freedesktop_dbus", delay=0.2)



# from bt import backend_to_check, gnome, kde

# "The process is not authorized to take a screenshot"

# if kde():

#     def test_kwin_dbus():
#         assert not gnome()
#         backend_to_check("kwin_dbus")



import platform

import pyscreenshot
from pyscreenshot import backend_version


def print_name_version(name, version):
    s = "{:<20} {}".format(name, version)
    print(s)


def print_versions():
    print_name_version("python", platform.python_version())
    print_name_version("pyscreenshot", pyscreenshot.__version__)

    for name in pyscreenshot.backends():
        v = backend_version(name)
        if not v:
            v = ""
        print_name_version(name, v)


if __name__ == "__main__":
    print_versions()



from bt import backend_to_check, check_import

from pyscreenshot.util import platform_is_osx

# qt color problem on osx
if not platform_is_osx():
    if check_import("PyQt5.Qt"):
        if check_import("PyQt5.QtGui"):
            if check_import("PyQt5.QtWidgets"):

                def test_pyqt5():
                    backend_to_check("pyqt5")



import logging

from easyprocess import EasyProcess

from pyscreenshot.plugins.backend import CBackend
from pyscreenshot.tempexport import RunProgError, read_func_img
from pyscreenshot.util import extract_version

log = logging.getLogger(__name__)


# wikipedia:  https://en.wikipedia.org/wiki/Xwd
#       xwd | xwdtopnm | pnmtopng > Screenshot.png
# xwdtopnm is buggy:  https://bugs.launchpad.net/ubuntu/+source/netpbm-free/+bug/1379480
# solution : imagemagick convert
#       xwd -root -display :0 | convert xwd:- file.png
# TODO: xwd sometimes grabs the wrong window so this backend will be not added now
PROGRAM = "xwd"


def read_xwd_img():
    def run_prog(fpng, bbox=None):
        fxwd = fpng + ".xwd"
        pxwd = EasyProcess([PROGRAM, "-root", "-out", fxwd])
        pxwd.call()
        if pxwd.return_code != 0:
            raise RunProgError(pxwd.stderr)

        pconvert = EasyProcess(["convert", "xwd:" + fxwd, fpng])
        pconvert.call()
        if pconvert.return_code != 0:
            raise RunProgError(pconvert.stderr)

    im = read_func_img(run_prog)
    return im


class XwdWrapper(CBackend):
    name = "xwd"
    is_subprocess = True

    def grab(self, bbox=None):
        im = read_xwd_img()
        if bbox:
            im = im.crop(bbox)
        return im

    def backend_version(self):
        return extract_version(EasyProcess([PROGRAM, "-version"]).call().stdout)



"""
import os

# import tkinter
import pygame
from easyprocess import EasyProcess

from pyscreenshot.util import platform_is_linux, platform_is_osx, platform_is_win


def display_size_x():
    # http://www.cyberciti.biz/faq/how-do-i-find-out-screen-resolution-of-my-linux-desktop/
    # xdpyinfo  | grep 'dimensions:'
    screen_width, screen_height = 0, 0
    if not os.environ.get("DISPLAY"):
        raise ValueError("missing DISPLAY variable")
    xdpyinfo = EasyProcess("xdpyinfo")
    xdpyinfo.enable_stdout_log = False
    if xdpyinfo.call().return_code != 0:
        raise ValueError("xdpyinfo error: %s" % xdpyinfo)
    for x in xdpyinfo.stdout.splitlines():
        if "dimensions:" in x:
            screen_width, screen_height = map(int, x.strip().split()[1].split("x"))

    return screen_width, screen_height


def display_size_osx():
    from Quartz import CGDisplayBounds
    from Quartz import CGMainDisplayID

    mainMonitor = CGDisplayBounds(CGMainDisplayID())
    return int(mainMonitor.size.width), int(mainMonitor.size.height)


def display_size_win():
    from win32api import GetSystemMetrics

    return int(GetSystemMetrics(0)), int(GetSystemMetrics(1))


# def display_size_tk():
#     root = tkinter.Tk()
#     w, h = root.winfo_screenwidth(), root.winfo_screenheight()
#     return (w, h)


def display_size_pygame():
    infoObject = pygame.display.Info()
    return infoObject.current_w, infoObject.current_h


def display_size():
    if platform_is_osx():
        w, h = display_size_osx()

    if platform_is_win():
        w, h = display_size_win()

    if platform_is_linux():
        w, h = display_size_x()

    # pgw, pgh = display_size_pygame()
    # assert pgw == w
    # assert pgh == h
    return w, h
"""



from bt import backend_to_check, gnome, gnome_version, kde

if gnome():
    # GNOME Shell 41.0 disallowed unrestricted access to the screenshot API
    if gnome_version()[0] < 41:

        def test_gnome_dbus():
            assert not kde()
            backend_to_check("gnome_dbus")



import io

from PIL import Image

# def _coder(im):
#     if im:
#         data = {
#             'pixels': im.tobytes(),
#             'size': im.size,
#             'mode': im.mode,
#         }
#         return data
#
#
# def _decoder(data):
#     if data:
#         im = Image.frombytes(data['mode'], data['size'], data['pixels'])
#         return im


def _coder(im):
    if im:
        b = io.BytesIO()
        im.save(b, format="png")
        data = b.getvalue()
        return data


def _decoder(data):
    if data:
        b = io.BytesIO(data)
        im = Image.open(b)
        return im


codec = (_coder, _decoder)



import logging
import os
import sys
from time import sleep

import fillscreen
from config import bbox_ls
from double_disp import check_double_disp
from easyprocess import EasyProcess
from entrypoint2 import entrypoint
from image_debug import ImageDebug
from PIL import Image, ImageChops

import pyscreenshot
from pyscreenshot.util import platform_is_linux

# backend tester (bt)

imdbg = ImageDebug()


def check_ref(backend, bbox, childprocess, refimgpath):
    im = pyscreenshot.grab(bbox=bbox, backend=backend, childprocess=childprocess)

    img_ref = Image.open(refimgpath)
    logging.debug("ref full getextrema: %s", img_ref.getextrema())
    # if not bbox:
    #     bbox = img_ref.getbbox()
    #     # rm tkinter border
    #     bw = 1
    #     bbox = (bbox[0] + bw, bbox[1] + bw, bbox[2] - bw, bbox[3] - bw)
    #     img_ref = img_ref.crop(bbox)
    #     im = im.crop(bbox)
    # elif bbox:
    if bbox:
        img_ref = img_ref.crop(bbox)

    img_ref = img_ref.convert("RGB")
    logging.debug("ref  getextrema: %s", img_ref.getextrema())
    im = im.convert("RGB")
    logging.debug("shot getextrema: %s", im.getextrema())

    assert "RGB" == img_ref.mode
    assert "RGB" == im.mode

    imdbg.img_debug(img_ref, "ref" + str(bbox))
    imdbg.img_debug(im, str(backend) + str(bbox))

    img_diff = ImageChops.difference(img_ref, im)
    ex = img_diff.getextrema()
    logging.debug("diff getextrema: %s", ex)
    diff_bbox = img_diff.getbbox()
    if diff_bbox:
        imdbg.img_debug(img_diff, "img_diff" + str(diff_bbox))
    # if (
    #     platform_is_osx()
    #     and backend
    #     and backend in ["pyqt", "pyqt5", "pyside", "pyside2"]
    # ):
    #     color_diff_max = max([b for (_, b) in ex])
    #     assert color_diff_max < 70
    # else:
    if diff_bbox is not None:
        print(
            "different image data %s bbox=%s extrema:%s diff_bbox=%s"
            % (backend, bbox, ex, diff_bbox)
        )
    assert diff_bbox is None


def backend_ref(backend, childprocess=True, refimgpath="", delay=0):
    for bbox in bbox_ls:
        print("bbox: {}".format(bbox))
        print("backend: %s" % backend)
        check_ref(backend, bbox, childprocess, refimgpath)
        if delay:
            sleep(delay)


def backend_to_check(backend, delay=0):
    refimgpath = fillscreen.init()
    backend_ref(backend, childprocess=True, refimgpath=refimgpath, delay=delay)

    # childprocess=False is tested in a subprocess for isolation
    cmd = [
        sys.executable,
        __file__.rsplit(".", 1)[0] + ".py",
        backend if backend else "",
        refimgpath,
        str(delay),
        "--debug",
    ]
    p = EasyProcess(cmd).call()
    assert p.return_code == 0

    if platform_is_linux() and prog_check(["Xvfb", "-help"]):
        check_double_disp(backend)


@entrypoint
def main(backend, refimgpath: str, delay: float):
    if not backend:
        backend = None
    # delay = int(delay)
    backend_ref(backend, childprocess=False, refimgpath=refimgpath, delay=delay)


def kde():
    XDG_CURRENT_DESKTOP = os.environ.get("XDG_CURRENT_DESKTOP")
    if XDG_CURRENT_DESKTOP:
        return "kde" in XDG_CURRENT_DESKTOP.lower()


def gnome():
    XDG_CURRENT_DESKTOP = os.environ.get("XDG_CURRENT_DESKTOP")
    if XDG_CURRENT_DESKTOP:
        return "gnome" in XDG_CURRENT_DESKTOP.lower()


def gnome_version():
    p = EasyProcess(["gnome-shell", "--version"]).call()
    # GNOME Shell 41.3
    if p.return_code != 0:
        return
    # 41.3
    s = p.stdout.split()[-1]
    return [int(x) for x in s.split(".")]


def kde_version():
    p = EasyProcess(["plasmashell", "--version"]).call()
    # plasmashell 5.14.5
    if p.return_code != 0:
        return
    # 5.14.5
    s = p.stdout.split()[-1]
    return [int(x) for x in s.split(".")]


def check_import(module):
    found = False
    # try:
    #     __import__(module)

    #     ok = True
    # except ImportError:
    #     pass

    import importlib

    try:
        spam_spec = importlib.util.find_spec(module)
        found = spam_spec is not None
        return found
    except ModuleNotFoundError:
        return False


def prog_check(cmd):
    try:
        if EasyProcess(cmd).call().return_code == 0:
            return True
    except Exception:
        return False



#!/usr/bin/env python3
import os
import os.path
import threading
import time
from pathlib import Path
from tempfile import TemporaryDirectory
from time import sleep

import fabric
import imagehash
import vagrant
from easyprocess import EasyProcess
from entrypoint2 import entrypoint
from PIL import Image

KBD_RIGHT = "e0 4d e0 cd".split()
KBD_SPACE = "39 b9".split()


def vbox_send_kbd(box, kbd):
    cmd = ["VBoxManage", "controlvm", box, "keyboardputscancode"] + kbd
    EasyProcess(cmd).call()


def close_confirm_wnd(box):
    vbox_send_kbd(box, KBD_RIGHT + KBD_RIGHT + KBD_SPACE)


def vbox_screenshot(box, bbox=None):
    with TemporaryDirectory(prefix="vbox_screenshot") as tmpdirname:
        filename = os.path.join(tmpdirname, "screenshot.png")

        cmd = ["VBoxManage", "controlvm", box, "screenshotpng", filename]
        EasyProcess(cmd).call()

        im = Image.open(filename)
        if bbox:
            im = im.crop(bbox)  # (left, upper, right, lower)
        return im


# from fabric.api import env, execute, task, run, sudo, settings


# pip3 install fabric vncdotool python-vagrant entrypoint2

DIR = Path(__file__).parent


class Options:
    halt = True
    recreate = True
    destroy = False
    fast = False
    create = False


def wrapcmd(cmd, guiproc):
    if guiproc:
        # copy env vars from graphical session
        cmd2 = f"pid=`pidof {guiproc}`"
        cmd2 += ";pid=$( cut -d ' ' -f 1 <<< $pid )"
        cmd2 += ";sudo cat /proc/$pid/environ | tr '\\0' '\\n' > /tmp/environ"
        cmd2 += ";export $(cat /tmp/environ)"
        cmd = cmd2 + f"; {cmd}"
    print(f"command: {cmd}")
    return cmd


def run_box(name, options, vagrantfile, cmds, guiproc):
    env = os.environ
    env["VAGRANT_VAGRANTFILE"] = str(DIR / vagrantfile)
    env["VAGRANT_DOTFILE_PATH"] = str(DIR / (".vagrant_" + vagrantfile))

    v = vagrant.Vagrant(env=env, quiet_stdout=False, quiet_stderr=False)
    status = v.status()
    state = status[0].state
    print(status)

    if options.destroy:
        v.destroy()
        return

    if options.halt:
        v.halt()  # avoid screensaver

    if state == "not_created":
        # install programs in box
        v.up()

        # restart box
        v.halt()

        if not options.fast:
            # go over first boot messages, tasks
            v.up()
            sleep(3 * 60)
            v.halt()

    if options.create:
        return

    event = threading.Event()

    auto_confirm_thread = None
    if name == "ubuntu.22.04":

        box = "pyscreenshot_" + name

        def auto_confirm():
            confirm_wnd_hash = "0000207e3e3e0000"
            bbox = (610, 120, 670, 150)
            while True:
                if event.is_set():
                    break
                img = vbox_screenshot(box, bbox)
                h = imagehash.average_hash(img)
                # print(h)
                if str(h) == confirm_wnd_hash:
                    # print("close_confirm_wnd")
                    close_confirm_wnd(box)
                time.sleep(1)

        auto_confirm_thread = threading.Thread(target=auto_confirm)

    try:
        v.up()

        with fabric.Connection(
            v.user_hostname_port(),
            connect_kwargs={
                "key_filename": v.keyfile(),
            },
        ) as conn:
            with conn.cd("c:/vagrant" if options.win else "/vagrant"):
                if not options.win:
                    if options.osx:
                        freecmd = "top -l 1 -s 0 | grep PhysMem"
                    else:  # linux
                        freecmd = "free -h"
                    cmds = (
                        [
                            freecmd,
                            "env | sort",
                            "gnome-shell --version || true",
                            "plasmashell --version || true",
                        ]
                        + cmds
                        + [freecmd]
                    )

                if guiproc:
                    pid = None
                    while not pid:
                        print(f"waiting for {guiproc}")
                        cmd = f"bash --login -c 'pidof {guiproc} || true'"
                        print(cmd)
                        pid = conn.run(cmd).stdout
                        sleep(1)
                    print(f"{guiproc} pid={pid}")
                sleep(5)

                if auto_confirm_thread:
                    auto_confirm_thread.start()

                print(f"collected commands: {cmds}")
                for cmd in cmds:
                    if options.recreate:
                        if "tox" in cmd:
                            # cmd += " -r"
                            cmd = cmd.replace("tox", "tox -r")
                    conn.run(wrapcmd(cmd, guiproc), echo=True, pty=True)
    finally:
        event.set()
        if auto_confirm_thread:
            auto_confirm_thread.join()
        if options.halt:
            v.halt()


config = {
    "ubuntu.server.22.04": (
        ["tox"],
        "",
    ),
    "lubuntu.20.04": (
        ["tox -e py3-desktop"],
        "qterminal",
    ),
    "lubuntu.22.04": (
        ["tox -e py3-desktop"],
        "qterminal",
    ),
    "lubuntu.22.10": (
        ["tox -e py3-desktop"],
        "qterminal",
    ),
    "xubuntu.20.04": (
        ["tox -e py3-desktop"],
        "xfdesktop",
    ),
    "xubuntu.22.04": (
        ["tox -e py3-desktop"],
        "xfdesktop",
    ),
    "xubuntu.22.10": (
        ["tox -e py3-desktop"],
        "xfdesktop",
    ),
    "kubuntu.20.04": (
        ["tox -e py3-desktop"],
        "konsole",
    ),
    "kubuntu.22.04": (
        ["tox -e py3-desktop-freedesktop"],
        "konsole",
    ),
    "kubuntu.22.10": (
        ["tox -e py3-desktop-freedesktop"],
        "konsole",
    ),
    "ubuntu.20.04": (
        ["tox -e py3-desktop"],
        "gnome-terminal-server",
    ),
    "ubuntu.22.04": (
        ["tox -e py3-desktop-freedesktop"],
        "gnome-terminal-server",
    ),
    "ubuntu.22.10": (
        ["tox -e py3-desktop-freedesktop"],
        "gnome-terminal-server",
    ),
    "ubuntu.22.04.sway": (
        ["tox -e py3-desktop"],
        "foot",
    ),
    # "debian11.gnome.wayland": (
    #     ["tox -e py3-desktop"],
    #     "gnome-terminal-server",
    # ),
    # "debian11.gnome.x11": (
    #     ["tox -e py3-desktop"],
    #     "gnome-terminal-server",
    # ),
    # "debian11.kde.wayland": (
    #     ["tox -e py3-desktop"],
    #     "konsole",
    # ),
    # "debian11.kde.x11": (
    #     ["tox -e py3-desktop"],
    #     "konsole",
    # ),
    # "debian10.gnome.wayland": (
    #     # ["bash -c 'tox -e py3-desktop'"],
    #     ["tox -e py3-desktop"],
    #     "gnome-terminal-server",
    # ),
    # "debian10.gnome.x11": (
    #     # ["bash -c 'tox -e py3-desktop'"],
    #     ["tox -e py3-desktop"],
    #     "gnome-terminal-server",
    # ),
    # "debian10.kde.wayland": (
    #     ["tox -e py3-desktop"],
    #     "konsole",
    # ),
    # "debian10.kde.x11": (
    #     ["tox -e py3-desktop"],
    #     "konsole",
    # ),
    # "osx.10.14": (
    #     ["bash --login -c 'python3 -m tox -e py3-osx'"],
    #     "Dock",
    # ),
    # "osx.10.15": (
    #     ["bash --login -c 'python3 -m tox -e py3-osx'"],
    #     "Dock",
    # ),
    # "win": ("Vagrantfile.win.rb", ["tox -e py3-win"], "",),
}


@entrypoint
def main(boxes="all", fast=False, destroy=False, create=False):
    options = Options()
    options.halt = not fast
    options.recreate = not fast
    options.destroy = destroy
    options.fast = fast
    options.create = create

    if boxes == "all":
        boxes = list(config.keys())
    else:
        boxes = boxes.split(",")

    for k, v in config.items():
        name = k
        vagrantfile = "Vagrantfile." + name + ".rb"
        cmds, guiproc = v[0], v[1]
        if name in boxes:
            options.win = k.startswith("win")
            options.osx = k.startswith("osx")
            print("----->")
            print("----->")
            print("-----> %s %s %s" % (name, vagrantfile, cmds))
            print("----->")
            print("----->")
            try:
                run_box(name, options, vagrantfile, cmds, guiproc)
            finally:
                print("<-----")
                print("<-----")
                print("<----- %s %s %s" % (name, vagrantfile, cmds))
                print("<-----")
                print("<-----")



from bt import backend_to_check, prog_check

from pyscreenshot.util import use_x_display

if not use_x_display():
    if prog_check(["grim", "-h"]):

        def test_grim():
            backend_to_check("grim")



from bt import backend_to_check, check_import

from pyscreenshot import FailedBackendError
from pyscreenshot.util import (
    platform_is_linux,
    platform_is_osx,
    platform_is_win,
    use_x_display,
)

try:
    from Xlib import display
except ImportError:
    display = None


def missing_RANDR():
    # https://github.com/python-xlib/python-xlib/blob/master/examples/xrandr.py#L44
    if platform_is_osx():
        return False
    if platform_is_win():
        return False
    if not display:
        return False
    disp = display.Display()
    return not disp.has_extension("RANDR")


ok = False
if check_import("mss"):
    if platform_is_osx() and not use_x_display():
        ok = True
    if platform_is_linux() and use_x_display():
        ok = True
    if platform_is_win():
        ok = True

if ok:

    def test_mss():
        if missing_RANDR():
            try:
                backend_to_check("mss")
            except FailedBackendError:
                pass
        else:
            backend_to_check("mss")



from easyprocess import EasyProcess

from pyscreenshot.plugins.backend import CBackend
from pyscreenshot.tempexport import read_prog_img
from pyscreenshot.util import extract_version

PROGRAM = "gnome-screenshot"

# https://gitlab.gnome.org/GNOME/gnome-screenshot/blob/master/src/screenshot-utils.c
# DBus is used for screenshot.
# If it doesn't succeed or $GNOME_SCREENSHOT_FORCE_FALLBACK is set then X DISPLAY is used.
# Flash effect! https://bugzilla.gnome.org/show_bug.cgi?id=672759


class GnomeScreenshotWrapper(CBackend):
    """Plugin for ``pyscreenshot`` that uses ``gnome-screenshot``
    https://git.gnome.org/browse/gnome-screenshot/

    This plugin can take screenshot when system is running Wayland.
    Info: https://bugs.freedesktop.org/show_bug.cgi?id=98672#c1
    """

    name = "gnome-screenshot"
    is_subprocess = True

    def grab(self, bbox=None):
        im = read_prog_img([PROGRAM, "-f"])
        if bbox:
            im = im.crop(bbox)
        return im

    def backend_version(self):
        p = EasyProcess([PROGRAM, "--version"])
        p.enable_stdout_log = False
        p.enable_stderr_log = False
        p.call()
        return extract_version(p.stdout.replace("-", " "))



import logging

from easyprocess import EasyProcess

from pyscreenshot.plugins.backend import CBackend
from pyscreenshot.tempexport import read_prog_img
from pyscreenshot.util import extract_version

log = logging.getLogger(__name__)


PROGRAM = "maim"


class MaimWrapper(CBackend):
    name = "maim"
    is_subprocess = True

    def grab(self, bbox=None):
        cmd = [PROGRAM, "--hidecursor"]
        if bbox:
            width = bbox[2] - bbox[0]
            height = bbox[3] - bbox[1]
            # https://github.com/naelstrof/maim/issues/119
            cmd += ["-g", "{}x{}+{}+{}".format(width, height, bbox[0], bbox[1])]
        im = read_prog_img(cmd)
        return im

    def backend_version(self):
        return extract_version(EasyProcess([PROGRAM, "--version"]).call().stdout)



import os.path

from setuptools import setup

NAME = "pyscreenshot"

# get __version__
__version__ = None
exec(open(os.path.join(NAME, "about.py")).read())
VERSION = __version__


URL = "https://github.com/ponty/pyscreenshot"
DESCRIPTION = "python screenshot"
LONG_DESCRIPTION = """The pyscreenshot module is obsolete in most cases.
It was created because PIL ImageGrab module worked on Windows only,
but now Linux and macOS are also supported.
There are some features in pyscreenshot which can be useful in special cases:
flexible backends, Wayland support, sometimes better performance, optional subprocessing.

The module can be used to copy the contents of the screen to a Pillow image memory using various back-ends.
Replacement for the ImageGrab Module.

Documentation: https://github.com/ponty/pyscreenshot/tree/"""
LONG_DESCRIPTION += VERSION

PACKAGES = [
    NAME,
    NAME + ".plugins",
    NAME + ".check",
    NAME + ".cli",
    NAME + ".examples",
]

classifiers = [
    # Get more strings from
    # http://www.python.org/pypi?%3Aaction=list_classifiers
    "License :: OSI Approved :: BSD License",
    "Natural Language :: English",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: 3.11",
]

install_requires = [
    "EasyProcess",
    "entrypoint2",
    "mss ; python_version > '3.4'",
    "jeepney ; python_version > '3.4' and platform_system == 'Linux'",
]

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/x-rst",
    python_requires=">=3.4",
    classifiers=classifiers,
    keywords="screenshot",
    author="ponty",
    # author_email='',
    url=URL,
    license="BSD",
    packages=PACKAGES,
    install_requires=install_requires,
    package_data={
        NAME: ["py.typed"],
    },
)



from bt import backend_to_check

from pyscreenshot.util import (
    platform_is_linux,
    platform_is_osx,
    platform_is_win,
    use_x_display,
)

ok = False
if platform_is_osx() and not use_x_display():
    ok = True
if platform_is_linux() and use_x_display():
    ok = True
if platform_is_win():
    ok = True

if ok:

    def test_pil():
        backend_to_check("pil")



import atexit
import logging
import sys
import tempfile
from os.path import dirname, join
from shutil import rmtree
from time import sleep

from easyprocess import EasyProcess
from PIL import Image, ImageChops

import pyscreenshot
from pyscreenshot.util import platform_is_osx, platform_is_win

# from image_debug import ImageDebug


log = logging.getLogger(__name__)

refimgpath = None


class FillscreenError(Exception):
    pass


def init():
    global refimgpath
    if not refimgpath:
        d = tempfile.mkdtemp(prefix="fillscreen")
        atexit.register(lambda: rmtree(d))
        refimgpath = join(d, "ref.bmp")

        # img_ref = generate_image()
        # img_ref.save(refimgpath)

        # fillscreen_tk = join(dirname(__file__), "fillscreen_tk.py")
        fillscreen_pygame = join(dirname(__file__), "fillscreen_pygame.py")
        python = sys.executable
        cmd = [python, fillscreen_pygame, "--image", refimgpath]
        cmd += ["--debug"]
        # if platform_is_win():
        #     cmd = [
        #         "C:\\Program Files (x86)\\FastStone Image Viewer\\FSViewer.exe",
        #         refimgpath,
        #     ]
        # else:
        #     cmd = [
        #         "pqiv",
        #         "--fullscreen",
        #         "--hide-info-box",
        #         "--disable-scaling",
        #         refimgpath,
        #     ]
        # if platform_is_osx():
        #     cmd = [
        #         "pqiv",
        #         "--fullscreen",
        #         "--hide-info-box",
        #         "--disable-scaling",
        #         refimgpath,
        #     ]
        # else:
        #     cmd = [python, fillscreen_pygame, "--image", refimgpath]
        proc = EasyProcess(cmd).start()
        atexit.register(proc.stop)
        log.debug(refimgpath)

        # sleep(30)
        # imdbg = ImageDebug()
        t = 0
        while True:
            sleep(2)
            t += 2
            if t > 90:
                raise FillscreenError("fillscreen timeout: %s" % t)

            if not proc.is_alive():
                raise FillscreenError("fillscreen stopped: %s" % proc)

            try:
                img_ref = Image.open(refimgpath)
            except FileNotFoundError:
                continue
            im = pyscreenshot.grab()
            # imdbg.img_debug(im, "raw")
            im = im.convert("RGB")
            img_diff = ImageChops.difference(img_ref, im)
            ex = img_diff.getextrema()
            log.debug("diff getextrema: %s", ex)
            color_diff_max = max([b for (_, b) in ex])
            diff_bbox = img_diff.getbbox()
            if diff_bbox is not None:
                log.debug(
                    "different image data. color_diff_max:%s extrema:%s diff_bbox=%s"
                    % (color_diff_max, ex, diff_bbox)
                )
                # imdbg.img_debug(img_ref, "ref")
                # imdbg.img_debug(im, "got")
                # imdbg.img_debug(img_diff, "img_diff" + str(diff_bbox))

            if color_diff_max < 10:
                break

        # if the OS has color correction
        #  then the screenshot has slightly different color than the original image
        if platform_is_win() or platform_is_osx():
            refimgpath = refimgpath + ".pil.png"
            im = pyscreenshot.grab(backend="pil")
            im.save(refimgpath)
            log.debug("%s saved", refimgpath)

    return refimgpath



# this test should come after backend tests
import os

from pyscreenshot.util import run_mod_as_subproc


def test_speedtest():
    assert run_mod_as_subproc("pyscreenshot.check.speedtest").return_code == 0
    assert (
        run_mod_as_subproc(
            "pyscreenshot.check.speedtest", ["--childprocess", "0"]
        ).return_code
        == 0
    )
    assert (
        run_mod_as_subproc(
            "pyscreenshot.check.speedtest", ["--childprocess", "1"]
        ).return_code
        == 0
    )


def test_print_versions():
    assert run_mod_as_subproc("pyscreenshot.check.versions").return_code == 0


def test_print_versions_no_path():
    path = os.environ["PATH"]
    os.environ["PATH"] = "xxx"
    try:
        assert run_mod_as_subproc("pyscreenshot.check.versions").return_code == 0
    finally:
        os.environ["PATH"] = path



import logging
from typing import List, Optional, Tuple

from PIL import Image

from pyscreenshot.about import __version__
from pyscreenshot.childproc import childprocess_backend_version
from pyscreenshot.loader import FailedBackendError, backend_dict, backend_grab

ADDITIONAL_IMPORTS = [FailedBackendError]

log = logging.getLogger(__name__)
log.debug("version=%s", __version__)


def grab(
    bbox: Optional[Tuple[int, int, int, int]] = None,
    childprocess: bool = True,
    backend: Optional[str] = None,
) -> "Image":
    """Copy the contents of the screen to PIL image memory.

    :param bbox: optional bounding box (x1,y1,x2,y2)
    :param childprocess: run back-end in new process using popen. (bool)
        This isolates back-ends from each other and from main process.
        Leave it as it is (True) to have a safe setting.
        Set it False to improve performance, but then conflicts are possible.
    :param backend: back-end can be forced if set (examples:scrot, wx,..),
                    otherwise back-end is automatic
    """
    if bbox:
        x1, y1, x2, y2 = bbox
        if x2 <= x1:
            raise ValueError("bbox x2<=x1")
        if y2 <= y1:
            raise ValueError("bbox y2<=y1")
    return backend_grab(backend, bbox, childprocess)


def backends() -> List[str]:
    """Back-end names as a list.

    :return: back-ends as string list
    """
    return list(backend_dict.keys())


def backend_version(backend: str) -> str:
    """Back-end version.

    :param backend: back-end (examples:scrot, wx,..)
    :return: version as string
    """
    return childprocess_backend_version(backend)



import logging
import os
from tempfile import TemporaryDirectory

from pyscreenshot.err import FailedBackendError
from pyscreenshot.imcodec import codec
from pyscreenshot.util import run_mod_as_subproc

log = logging.getLogger(__name__)


def childprocess_backend_version(backend):
    p = run_mod_as_subproc("pyscreenshot.cli.print_backend_version", [backend])
    if p.return_code != 0:
        log.warning(p)
        raise FailedBackendError(p)

    return p.stdout


def childprocess_grab(backend, bbox):
    with TemporaryDirectory(prefix="pyscreenshot") as tmpdirname:
        filename = os.path.join(tmpdirname, "screenshot.png")
        cmd = ["--filename", filename]
        if bbox:
            x1, y1, x2, y2 = map(str, bbox)
            bbox = ":".join(map(str, (x1, y1, x2, y2)))
            cmd += ["--bbox", bbox]
        if backend:
            cmd += ["--backend", backend]
        if log.isEnabledFor(logging.DEBUG):
            cmd += ["--debug"]

        p = run_mod_as_subproc("pyscreenshot.cli.grab", cmd)
        if p.return_code != 0:
            # log.debug(p)
            raise FailedBackendError(p)

        data = open(filename, "rb").read()
        data = codec[1](data)
        return data
